# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Describes a strong typing system that can then be extracted from a structured
object.
"""

from nr.collections import abc
from nr.databind.core import IDataType, InvalidTypeDefinitionError
from nr.interface import implements, override
from nr.pylang.utils import classdef
from nr.pylang.utils.typing import is_generic, get_generic_args

import datetime
import decimal
import six
import typing


@implements(IDataType)
class OptionalType(object):
  """
  Represents an optional type (value for this type can be of the wrapped
  type or #None).
  """

  classdef.comparable(['value_type'])

  def __init__(self, value_type):  # type: (IDataType) -> None
    self.value_type = translate_type_def(value_type)

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if is_generic(py_type_def, typing.Optional):
      value_type = recursive(get_generic_args(py_type_def)[0])
      return cls(value_type)
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    if py_value is None:
      return None
    return self.value_type.isinstance_check(py_value, strict, coerce)


@implements(IDataType)
class AnyType(object):
  """
  Represents an arbitrary type. The "Any" datatype is simply ignored by
  converters.
  """

  classdef.comparable([])

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if py_type_def is object or py_type_def is typing.Any:
      return cls()
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    return py_value


@implements(IDataType)
class BooleanType(object):

  classdef.comparable(['strict'])

  def __init__(self, strict=True):
    self.strict = strict

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if py_type_def is bool:
      return BooleanType()
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    if (self.strict or strict) and not isinstance(py_value, bool):
      raise TypeError('expected bool')
    return bool(py_value) if coerce else py_value


@implements(IDataType)
class StringType(object):

  classdef.comparable(['strict'])

  def __init__(self, strict=True):
    self.strict = strict

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if py_type_def is str:
      return StringType()
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    if isinstance(py_value, six.string_types):
      return py_value
    else:
      if self.strict or strict:
        raise TypeError('expected string, got {}'.format(
          type(py_value).__name__))
      return str(py_value) if coerce else py_value


@implements(IDataType)
class BinaryType(object):

  classdef.comparable([])

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if py_type_def is bytes:
      return BinaryType()
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    if isinstance(py_value, six.binary_type):
      return py_value
    else:
      raise TypeError('expected binary, got {}'.format(type(py_value).__name__))


@implements(IDataType)
class IntegerType(object):

  classdef.comparable(['strict'])

  def __init__(self, strict=True):
    self.strict = strict

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if py_type_def is int:
      return IntegerType()
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    if isinstance(py_value, six.integer_types):
      return py_value
    else:
      if self.strict or strict:
        raise TypeError('expected int, got {}'.format(type(py_value).__name__))
      return int(py_value) if coerce else py_value


@implements(IDataType)
class DecimalType(object):
  """
  Represents a decimal value, backed by a float or [[decimal.Decimal]] object.
  If the selected Python type is Decimal, it will always accept strings which
  will be coerced to the correct type.

  If the selected type is float, it will only accept a string as input if
  *strict* is set to False.
  """

  classdef.comparable(['python_type', 'decimal_context', 'strict'])
  classdef.repr(['python_type', 'decimal_context', 'strict'])

  def __init__(self, python_type, decimal_context=None, strict=True):
    if python_type not in (float, decimal.Decimal):
      raise ValueError('python_type must be float or decimal.Decimal, got {!r}'
                       .format(python_type))
    if python_type is not decimal.Decimal and decimal_context:
      raise ValueError('decimal_context can only be used if python_type is '
                       'decimal.Decimal, but got {!r}'.format(python_type))
    self.python_type = python_type
    self.decimal_context = decimal_context
    self.strict = strict
    if self.python_type is decimal.Decimal:
      self.accepted_input_types = (decimal.Decimal, six.string_types)
      if not self.strict:
        self.accepted_input_types += (six.integer_types, float)
    else:
      self.accepted_input_types = (six.integer_types, float)
      if not self.strict:
        self.accepted_input_types += (six.string_types,)

  def coerce(self, value):
    if self.python_type is decimal.Decimal:
      return decimal.Decimal(value, self.decimal_context)
    elif self.python_type is float:
      return float(value)
    else:
      raise RuntimeError('python_type is invalid: {!r}'.format(
        self.python_type))

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if py_type_def in (float, decimal.Decimal):
      return DecimalType(py_type_def)
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    if not isinstance(py_value, self.accepted_input_types):
      raise TypeError('expected {}'.format(
        '|'.join(x.__name__ for x in self.accepted_input_types)))
    return self.coerce(py_value) if coerce else py_value


@implements(IDataType)
class CollectionType(object):
  """
  Represents a collection of elements. The *py_type* represents the type for
  representing the collection in Python. It may also be a function that
  processes the returned list.
  """

  classdef.comparable(['item_type', 'py_type'])

  def __init__(self, item_type, py_type=list):
    self.item_type = item_type
    self.py_type = py_type

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    from .collection import Collection
    # []
    if isinstance(py_type_def, list) and len(py_type_def) == 0:
      return cls(AnyType())
    # [<type>]
    elif isinstance(py_type_def, list) and len(py_type_def) == 1:
      return cls(recursive(py_type_def[0]))
    # list
    elif py_type_def is list:
      return cls(AnyType())
    # set
    elif py_type_def is set:
      return cls(AnyType(), py_type=set)
    # typing.List
    elif is_generic(py_type_def, typing.List):
      item_type_def = get_generic_args(py_type_def)[0]
      if isinstance(item_type_def, typing.TypeVar):
        item_type_def = object
      return cls(recursive(item_type_def))
    # typing.Set
    elif is_generic(py_type_def, typing.Set):
      item_type_def = get_generic_args(py_type_def)[0]
      if isinstance(item_type_def, typing.TypeVar):
        item_type_def = object
      return cls(recursive(item_type_def), py_type=set)
    # nr.databind.core.collection.Collection
    elif isinstance(py_type_def, type) and issubclass(py_type_def, Collection):
      return py_type_def.datatype
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    if isinstance(py_value, six.string_types) \
        or not isinstance(py_value, abc.Iterable):
      raise TypeError('expected a collection type, got {}'.format(
        type(py_value).__name__))
    if coerce and (not isinstance(self.py_type, type) or
        not isinstance(py_value, self.py_type)):
      py_value = self.py_type(py_value)
    return py_value


@implements(IDataType)
class ObjectType(object):
  """
  Represents an object (ie. a dictionary in Python lingo).
  """

  classdef.comparable(['value_type'])

  def __init__(self, value_type, py_type=dict):
    assert IDataType.provided_by(value_type), value_type
    self.value_type = value_type
    self.py_type = py_type

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    # {}
    if isinstance(py_type_def, dict) and len(py_type_def) == 0:
      return cls(AnyType())
    # {<type>}
    elif isinstance(py_type_def, set) and len(py_type_def) == 1:
      return cls(recursive(next(iter(py_type_def))))
    # {'value_type': <type>}
    elif isinstance(py_type_def, dict) and len(py_type_def) == 1 and \
        'value_type' in py_type_def:
      return cls(recursive(py_type_def['value_type']))
    # dict
    elif py_type_def is dict:
      return cls(AnyType())
    # typing.Dict
    elif is_generic(py_type_def, typing.Dict):
      key_type_def, value_type_def = get_generic_args(py_type_def)
      if not isinstance(key_type_def, typing.TypeVar) and key_type_def is not str:
        raise InvalidTypeDefinitionError(py_type_def)
      if isinstance(value_type_def, typing.TypeVar):
        value_type_def = object
      return cls(recursive(value_type_def))
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    if not isinstance(py_value, abc.Mapping):
      raise TypeError('expected a mapping, got {!r}'.format(type(py_value).__name__))
    if coerce and (not isinstance(self.py_type, type) or
        not isinstance(py_value, self.py_type)):
      py_value = self.py_type(py_value)
    return py_value


@implements(IDataType)
class PythonClassType(object):
  """ A wrapper for Python type object's when they don't fit into any of the
  other types. The adapter for this datatype takes the lowest priority and
  only applies to custom defined types (not built-in types). """

  classdef.comparable(['cls', 'decorations'])
  classdef.repr(['cls', 'decorations'])

  def __init__(self, cls, *decorations):
    assert isinstance(cls, type)
    self.cls = cls
    self.decorations = list(decorations)

  @override
  @classmethod
  def priority(cls):
    return -1000

  @override
  def get_decorations(self):
    return self.decorations

  @override
  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if isinstance(py_type_def, type) and py_type_def.__module__ not in (
        'builtin', '__builtin__', 'typing') and not IDataType.implemented_by(py_type_def):
      return cls(py_type_def)
    raise InvalidTypeDefinitionError(py_type_def)

  @override
  def isinstance_check(self, py_value, strict, coerce):
    if not isinstance(py_value, self.cls):
      raise TypeError('expected {} instance, got {}'.format(
        self.cls.__name__, type(py_value).__name__))
    return py_value

  @classmethod
  def make_check(cls, py_type): # type: (Type) -> Callable[bool, [IDataType]]
    """ Returns a test function that accepts an #IDataType instance and
    returns #True if it is a #PythonClassType of the specified *py_type*. """

    def check(datatype):
      return isinstance(datatype, PythonClassType) and issubclass(datatype.cls, py_type)
    check.__qualname__ = 'PythonClassType.of.{}.check'.format(py_type.__name__)
    return check


@implements(IDataType)
class MultiType(object):
  """ Represents a collection of datatypes. Uses the first type of the list
  of types that successfully serializes/deserializes. Multi types can be
  defined conveniently using tuples. """

  classdef.comparable(['types'])

  def __init__(self, *types):
    if len(types) == 1:
      types = types[0]
    self.types = [translate_type_def(x) for x in types]

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if isinstance(py_type_def, tuple):
      return MultiType([recursive(x) for x in py_type_def])
    elif is_generic(py_type_def, typing.Union):
      return MultiType(get_generic_args(py_type_def))
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    errors = []
    for datatype in self.types:
      try:
        return datatype.isinstance_check(py_value, strict, coerce)
      except TypeError as exc:
        errors.append(exc)
    raise TypeError(errors)


@implements(IDataType)
class ProxyType(object):
  """ Proxy for another datatype. Bind it to an actual type with #implementation(). """

  classdef.comparable(['wrapped_type'])
  classdef.repr(['wrapped_type'])

  def __init__(self, wrapped_type=None):  # type: (Optional[IDataType]) -> None
    self.wrapped_type = wrapped_type

  def implementation(self, py_type_def):  # type: (Any) -> Any
    """
    Sets #wrapped_type to *py_type_def* translated to an #IDataType with
    #translate_type_def() and returns *py_type_def*. Example:

    >>> from nr.databind.core import ProxyType, Struct
    >>> A = ProxyType()
    >>> @A.implementation
    ... class A(Struct):
    ...   parent = Field(A, default=None)
    ...   # ...
    """

    self.wrapped_type = translate_type_def(py_type_def)
    return py_type_def

  @staticmethod
  def deref(datatype):  # type: (IDataType) -> IDataType
    """
    Dereferences an IDataType object. This returns the wrapped type if *datatype*
    is a #ProxyType instance, otherwise *datatype* is returned unchaged.

    :raise RuntimeError: If *datatype* is a #ProxyType but #ProxyType.wrapped_type
      is not set.
    """

    if isinstance(datatype, ProxyType):
      if not datatype.wrapped_type:
        raise RuntimeError('ProxyType.wrapped_type is None')
      return datatype.wrapped_type
    return datatype

  @override
  def to_human_readable(self):
    if self.wrapped_type:
      return self.wrapped_type.to_human_readable()
    else:
      return '<unbound ProxyType>'

  @override
  def propagate_field_name(self, name):
    pass

  @override
  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    raise InvalidTypeDefinitionError(py_type_def)

  @override
  def isinstance_check(self, py_value, strict, coerce):
    return self.wrapped_type.isinstance_check(py_value, strict, coerce)


def translate_type_def(py_type_def, fallback=None):
  if IDataType.provided_by(py_type_def):
    return py_type_def
  elif isinstance(py_type_def, type) and IDataType.implemented_by(py_type_def):
    try:
      return py_type_def()
    except TypeError:
      raise InvalidTypeDefinitionError(py_type_def)
  for adapter in sorted(IDataType.implementations(), key=lambda x: -x.priority()):
    try:
      return adapter.from_typedef(translate_type_def, py_type_def)
    except InvalidTypeDefinitionError:
      pass
  if fallback:
    return fallback
  raise InvalidTypeDefinitionError(py_type_def)


__all__ = [
  'OptionalType',
  'AnyType',
  'BooleanType',
  'StringType',
  'BinaryType',
  'IntegerType',
  'DecimalType',
  'CollectionType',
  'ObjectType',
  'PythonClassType',
  'MultiType',
  'ProxyType',
  'translate_type_def',
]



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '0.0.20.post1'

from nr.pylang.utils import NotSet
import inspect
import nr.pylang.utils.classdef
import nr.interface
import string

### Error classes
#################

class SerializationError(Exception):

  def __init__(self, node, message=None):  # type: (Node, Optional[str])
    if not isinstance(node, Node):
      raise TypeError('node must be a Location object, got {}'.format(
        type(node).__name__))
    self.node = node
    self.message = message

  def __str__(self):
    result = 'at {} (datatype: {})'.format(self.node.locator, self.node.datatype)
    if self.node.context.filename:
      result = 'in "{}": '.format(self.node.context.filename) + result
    if self.message:
      result += ': ' + str(self.message)
    return result


class SerializationTypeError(SerializationError):

  def __init__(self, node, message=None):  # type: (Location, Optional[str])
    if message is None:
      expected = node.datatype.to_human_readable()
      got = type(node.value).__name__
      message = 'expected "{}", got "{}"'.format(expected, got)
    super(SerializationTypeError, self).__init__(node, message)


class SerializationValueError(SerializationError):
  pass


class InvalidTypeDefinitionError(Exception):

  def __init__(self, py_type_def):
    self.py_type_def = py_type_def

  def __str__(self):
    return repr(self.py_type_def)


### Interfaces
##############

class IDataType(nr.interface.Interface):
  """
  Interface for datatypes. A datatype usually has a one to one mapping
  with a Python type. The serialization/deserialization of the type to other
  realms is handled by the #IDeserializer and #ISerializer interfaces.

  Datatypes must be comparable. The default string representation is derived
  from the constructor arguments. Datatypes may also provide a more easily
  readable representation with #to_human_readable().

  A datatype also describes how it can be created from a more Pythonic
  description via the #from_typedef() function.

  To implement an #IDataType, it is recommendable to use #nr.pylang.utils.\\
  classdef.comparable() in order to implement `__hash__()`, `__eq__()` and
  `__ne__()` in.
  """

  nr.pylang.utils.classdef.comparable([])  # adds __hash__, __eq__, __ne__ to the interface

  @nr.interface.default
  @classmethod
  def priority(cls):
    return 0

  @nr.interface.default
  def get_decorations(self):  # type: () -> List[Decoration]
    return []

  @nr.interface.default
  def __str__(self):  # type: () -> str
    attrs = {k: getattr(self, k) for k in self.__comparable_properties__}
    return '{}({})'.format(
      type(self).__name__,
      ', '.join('{}={!r}'.format(k, v) for k, v in attrs.items()),
    )

  @nr.interface.default
  def __repr__(self):  # type: () -> str
    return str(self)

  @nr.interface.default
  def to_human_readable(self):  # type: () -> str
    return type(self).__name__

  @nr.interface.default
  def propagate_field_name(self, name):  # type: (str) -> None
    """ This method is called when the datatype instance is attached to a
    field in an object. The name of the field is passed to this method. This
    is used for the inline object definition. """

  @classmethod
  def from_typedef(cls, recursive, py_type_def):  # type: (Callable, Any) -> IDataType
    # raises: InvalidTypeDefinitionError
    """ Convert the datatype from a pure Python description. Raise an
    #InvalidTypeDefinitionError if the *py_type_def* cannot be translated to
    this datatype. """

  def isinstance_check(self, py_value, strict, coerce):  # type: (Any, bool, bool) -> Any
    """ This method returns *py_value*, or an adaptation of *py_value*, if it
    matches the datatype. If it doesn't, a [[TypeError]] with the reason is
    raised.

    raises TypeError: If the *py_value* doesn't match this datatype. """


class IDeserializer(nr.interface.Interface):
  " Interface for deserializing values of a given #IDataType. "

  def deserialize(self, mapper, node):  # type: (ObjectMapper, Node) -> Any
    pass


class ISerializer(nr.interface.Interface):
  " Interface for serializing values of a given #IDataType. "

  def serialize(self, mapper, node):  # type: (ObjectMapper, Node) -> Any
    pass


### Supporting classes for (de-) serialization.
###############################################

class Locator(object):
  """
  A locator is basically a wrapper for a list of strings or integer indices
  that represents the location of an element in an object tree.

  Locator objects are immutable.
  """

  _WHITELISTED_KEY_CHARACTERS = string.ascii_letters + string.digits + '_-'

  def __init__(self, items):
    self._items = items

  __iter__ = lambda self: iter(self._items)
  __len__ = lambda self: len(self._items)
  __getitem__ = lambda self, index: self._items[index]
  __repr__ = lambda self: 'Locator({})'.format(str(self))
  __bool__ = lambda self: bool(self._items)
  __nonzero__ = __bool__

  def __str__(self):
    parts = ['$']
    for item in self._items:
      if isinstance(item, int):
        parts.append('[' + str(item) + ']')
      else:
        item = str(item)
        if '"' in item:
          item = item.replace('"', '\\"')
        if any(c not in self._WHITELISTED_KEY_CHARACTERS for c in item):
          item = '"' + item + '"'
        parts.append('.' + item)
    return ''.join(parts)

  def access(self, root):  # type: (Union[List, Dict]) -> Any
    """
    Accesses the value represented by this Locator object starting from *root*.

    >>> root = {'values': [1, 2, 3]}
    >>> from nr.databind.core.context import Locator
    >>> Locator(['values', 1]).access(root)
    2

    :param root: The object to start indexing from.
    :raises KeyError: If an item in an object cannot be accessed.
    :raises IndexError: If an item in a list cannot be accessed.
    ``` """

    value = root
    for item in self._items:
      try:
        value = value[item]
      except KeyError as exc:
        raise KeyError(str(self))
      except IndexError as exc:
        raise IndexError('{} at {}'.format(exc, self))

    return value

  def append(self, item):  # type: (Union[str, int]) -> Locator
    " Creates a copy of this locator and appends *item* to it. "

    items = list(self._items)
    items.append(item)
    return Locator(items)

  def pop(self, item):  # type: () -> Locator
    " Creates a copy of this locator and removes the last item from it. "

    items = list(self._items)
    items.pop()
    return Locator(items)


class Context(object):
  """
  A Context object contains all information that is mostly constant during the
  same (de-) serialization process.

  :param parent: The parent Context object. This is usually None unless a (de-)
    serializer decides to create a new sub-context.
  :param filename: The name of the file from which the data for the (de-)
    serialization originates. This is useful in case of an exception which can
    then include this filename for informational purposes.
  :param decorations: A list of decorations that apply to the context.
  """

  def __init__(self, parent, filename, decorations=None):
    # type: (Optional[Context], Optional[str], Optional[List[Decoration]]) -> None
    if parent is not None and not isinstance(parent, Context):
      raise TypeError('parent must be None or Context, got "{}"'.format(type(parent).__name__))
    if filename is not None and not isinstance(filename, str):
      raise TypeError('filename must be None or str, got "{}"'.format(type(filename).__name__))
    self.parent = parent
    self.filename = filename
    self.decorations = decorations or []


class Node(object):
  """
  A Node represents a value in a tree during (de-) serialization.

  :param parent: The parent node which represents the parent value.
  :param context: The Context object that is used for this node's processing.
  :param locator: A Locator object that describes the location of the node in
    the tree.
  :param datatype: The expected datatype of the node.
  :param value: The value of the node that is to be (de-) serialized.
  :param decorations: A list of decorations that apply to the node.
  """

  def __init__(self, parent, context, locator, datatype, value, decorations=None, prev_sibling=None):
    # type: (Optional[Node], Context, Locator, IDataType, Any, Optional[List[Decoration]]) -> None
    datatype = translate_type_def(datatype) if datatype is not None else None
    if parent is not None and not isinstance(parent, Node):
      raise TypeError('parent must be None or Node, got "{}"'.format(type(parent).__name__))
    if prev_sibling is not None and not isinstance(prev_sibling, Node):
      raise TypeError('prev_sibling must be None or Node, got "{}"'.format(type(prev_sibling).__name__))
    if not isinstance(context, Context):
      raise TypeError('context must be Context, got "{}"'.format(type(context).__name__))
    if not isinstance(locator, Locator):
      raise TypeError('locator must be Locator, got "{}"'.format(type(locator).__name__))
    if datatype is not None and not IDataType.provided_by(datatype):
      raise TypeError('datatype must be IDataType, got "{}"'.format(type(datatype).__name__))
    self.parent = parent
    self.prev_sibling = prev_sibling
    self.context = context
    self.locator = locator
    self.datatype = ProxyType.deref(datatype)
    self.value = value
    self.decorations = decorations or []
    self.result = None
    self.unknowns = set()
    collect = get_decoration(Collect, context.decorations)
    if collect:
      collect.nodes.append(self)

  def __repr__(self):
    return 'Node(locator={!r}, datatype={!r})'.format(self.locator, self.datatype)

  def type_error(self, message=None):  # type: (Optional[str]) -> SerializationTypeError
    " A helper function to create a #SerializationTypeError object. "

    return SerializationTypeError(self, message)

  def value_error(self, message=None):  # type: (Optional[str]) -> SerializationValueError
    " A helper function to create a #SerializationValueError object. "

    return SerializationValueError(self, message)

  def get_decoration(self, decoration_cls):  # type: (Type[Decoration]) -> Optional[Decoration]
    """
    Return the first decoration found for the *decoration_cls* in the Context's
    decorations, the Node's decorations, or, if applicable, the classes'
    decorations.
    """

    decoration = get_decoration(decoration_cls, self.context.decorations,
      self.decorations, self.datatype.get_decorations())
    if not decoration and issubclass(decoration_cls, ClassDecoration):
      if isinstance(self.datatype, StructType):
        decoration = decoration_cls.get(self.datatype.struct_cls)
      elif isinstance(self.datatype, PythonClassType):
        decoration = decoration_cls.get(self.datatype.cls)

    return decoration

  def get_local_decoration(self, decoration_cls):  # type: (Type[Decoration]) -> Optional[Decoration]
    """
    Return the first decoration found for the *decoration_cls* in the Node's
    decorations.
    """

    return get_decoration(decoration_cls, self.decorations)

  def make_child(self, key, datatype, value, context=None, decorations=None):
    return Node(self, context or self.context, self.locator.append(key),
                datatype, value, decorations)

  def replace(self, locator=NotSet, datatype=NotSet, value=NotSet, context=NotSet):
    locator = self.locator if locator is NotSet else locator
    datatype = self.datatype if datatype is NotSet else datatype
    value = self.value if value is NotSet else value
    context = self.context if context is NotSet else context
    return Node(None, context, locator, datatype, value, self.decorations, prev_sibling=self)

  @property
  def key(self):  # type: () -> Union[None, int, str]
    if self.locator:
      return self.locator[-1]
    else:
      return None


class Decoration(object):
  """
  Base class for decorations. Decorations can be attached to objects in order
  to customize the (de-) serialization process or the behavior of related types.

  * A #Context can have a list of decorations that influence the (de-) serialization
    behavior globally.
  * A #Node can have a list of decorations that influence the behavior (de-)
    serialization behavior locally.
  * #Field#s for #Struct#s can have decorations attached that influence the behavior
    for just the (de-) serialization for that field.
  * Classes can have decorations (in their `__decorations__` member) that may be
    respected during (de-) serialization.
  """


class SerializerCollection(object):
  """
  A collection of (de-) serializers. Serializers can be registered by association
  with the #IDataType that they are a (de-) serializer for or with a test callable.
  """

  def __init__(self, interface):
    self._interface = interface
    self._mapping = {}  # type: Dict[IDataType, interface]
    self._pytype_mapping = {}  # type: Dict[Type, interface]
    self._tests = []  # type: List[Tuple[Callable[[IDataType], Bool], interface]

  def register(self, arg, impl):
    """
    Registers a new #IDataType or test function along with  a(de-) serializer.
    """

    # Is this type applicable for the PythonClassType?
    is_pytype = False
    if isinstance(arg, type):
      try:
        PythonClassType.from_typedef(None, arg)
        is_pytype = True
      except InvalidTypeDefinitionError as exc:
        # It is not. :-)
        pass

    if not is_pytype and self._interface and not self._interface.provided_by(impl):
      raise TypeError('expected {} implementation, got {}'.format(
        self._interface.__name__, type(impl).__name__))

    if is_pytype:
      self._pytype_mapping[arg] = impl
    elif isinstance(arg, type) and (not self._interface or IDataType.implemented_by(arg)):
      self._mapping[arg] = impl
    elif inspect.isfunction(arg):
      self._tests.append((arg, impl))
    else:
      raise TypeError('expected IDataType or function, got {}'.format(type(arg).__name__))

  def match(self, datatype):
    for test_func, impl in self._tests:
      if test_func(datatype):
        return impl
    if isinstance(datatype, PythonClassType):
      try:
        return self._pytype_mapping[datatype.cls]
      except KeyError:
        pass
    try:
      return self._mapping[type(datatype)]
    except KeyError:
      pass
    return None


class Module(object):
  """
  A module is a collection of (de-) serializers. Modules can also be combined
  into one single module. Modules implement the #IDeserializeDispatcher and
  #ISerializeDispatcher interface.
  """

  def __init__(self, *submodules):
    self._serializers = SerializerCollection(ISerializer)
    self._deserializers = SerializerCollection(IDeserializer)
    self._submodules = []
    for module in submodules:
      self.register(module)

  def register(self, value, impl=None):
    if impl is None:
      if not isinstance(value, Module):
        raise TypeError('argument 1 must be a Module if argument 2 is omitted, got "{}"'
                        .format(type(value).__name__))
      self._submodules.append(value)
    else:
      valid = False
      if ISerializer.provided_by(impl):
        self._serializers.register(value, impl)
        valid = True
      if IDeserializer.provided_by(impl):
        self._deserializers.register(value, impl)
        valid = True
      if not valid:
        raise TypeError('argument 2 must implement ISerializer/IDeserializer, got "{}"'
                        .format(type(impl).__name__))

  def get_serializer(self, datatype):
    result = self._serializers.match(datatype)
    if result is None:
      for module in self._submodules:
        result = module.get_serializer(datatype)
        if result is not None:
          break
    return result

  def get_deserializer(self, datatype):
    result = self._deserializers.match(datatype)
    if result is None:
      for module in self._submodules:
        result = module.get_deserializer(datatype)
        if result is not None:
          break
    return result


class ObjectMapper(Module):
  """
  The ObjectMapper is an extension of the #Module class that provides additional
  functions to dispatch the (de-) serialization from an initial state.
  """

  def deserialize_node(self, node):  # type: (Node) -> Any
    deserializer = self.get_deserializer(node.datatype)
    if not deserializer:
      raise node.type_error('no deserializer was found for datatype {!r}'.format(node.datatype))
    node.result = deserializer.deserialize(self, node)
    return node.result

  def serialize_node(self, node):  # type: (Node) -> Any
    serializer = self.get_serializer(node.datatype)
    if not serializer:
      raise node.type_error('no serializer was found for datatype {!r}'.format(node.datatype))
    node.result = serializer.serialize(self, node)
    return node.result

  def deserialize_to_node(self, *args, **kwargs):  # type: (...) -> Node
    node = make_node(*args, **kwargs)
    self.deserialize_node(node)
    return node

  def serialize_to_node(self, *args, **kwargs):  # type: (...) -> Node
    node = make_node(*args, **kwargs)
    self.serialize_node(node)
    return node

  def deserialize(self, *args, **kwargs):  # type: (...) -> Any
    return self.deserialize_node(make_node(*args, **kwargs))

  def serialize(self, *args, **kwargs):  # type: (...) -> Any
    return self.serialize_node(make_node(*args, **kwargs))

  def cast(self, py_value, target_type):  # type: (Union[Struct, Collection], Type) -> Any
    """
    Cast *py_value* to *target_type* by serializing and deserializing the
    value.
    """

    data = self.serialize(py_value, type(py_value))
    return self.deserialize(data, target_type)


def make_node(value, datatype, filename=None, decorations=None):
  # type: (Any, Any, Optional[str], Optional[List[Decoration]]) -> (Context, Node)
  datatype = translate_type_def(datatype)
  context = Context(None, filename, decorations)
  node = Node(None, context, Locator([]), datatype, value)
  return node


### Feature imports
###################

from .collection import Collection
from .datatypes import *
from .decorations import *
from .struct import Field, Struct, StructType, make_struct, cast_struct
from .union import UnionType



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" This module implements common helpers for dealing with decorations. """

from nr.databind.core import Decoration, IDataType
from nr.databind.core.datatypes import IntegerType, ObjectType, StringType, translate_type_def
from nr.pylang.utils import classdef


# Base classes for decorations.

class FieldDecoration(Decoration):
  " Base class for decorations of fields. "

  def attached(self, field):  # type: (Field) -> None
    " Called when a field decoration is attached to a field. "

    pass


class ClassDecoration(Decoration):
  """
  Base class for decorations of classes. Class instances can be used as
  decorators to add themselves to the `__decorations__` member of the class.

  Class decorations by default are not inherited by subclasses but this behavior
  can be enabled by setting the #inheritable property to #True.
  """

  inheritable = False

  def __init__(self, inheritable=False):  # type: () -> None
    self.inheritable = inheritable

  def __call__(self, cls):  # type: (Type) -> Type
    """
    Add the decoration to *cls*. Creates a new `__decorations__` attribute
    if it does not exist on the class.
    """

    if '__decorations__' not in vars(cls):
      decorations = cls.__decorations__ = []
    else:
      decorations = cls.__decorations__
    decorations.append(self)
    return cls

  @classmethod
  def get(cls, decorated_cls):  # type: (Type) -> Optional[ClassDecoration]
    """
    Returns the first instance of *cls* found in the decorations for the type
    object specified with *decorated_cls*. If #inheritable is #True, the
    base classes of *decorated_cls* are searched as well.
    """

    if not isinstance(decorated_cls, type):
      raise TypeError('expected type instance, got {!r}'.format(type(decorated_cls).__name__))

    decoration = get_decoration(cls, vars(decorated_cls).get('__decorations__', []))
    if decoration is not None:
      return decoration

    for base in decorated_cls.__bases__:
      decoration = cls.get(base)
      if decoration is not None and decoration.inheritable:
        return decoration

    return None


class GlobalDecoration(Decoration):
  """
  Just a base class to indicate when a decoration can also be used globally.
  """


def get_decoration(decoration_cls, *decorations_lists):
  # type: (Type[Decoration], List[Decoration]) -> Optional[Decoration]
  """
  Given a #Decoration type object and zero or more lists of decorations, finds
  the first element that has the exact same type as *decoration_cls* and
  returns it.
  """

  for decorations in decorations_lists:
    for decoration in decorations:
      if type(decoration) is decoration_cls:
        return decoration

  return None


# Common concrete decorations.

class Collect(GlobalDecoration):
  """
  A global decoration that, if present, will be used to collect all nodes
  that are passed into #ObjectMapper.deserialize_node() or
  #ObjectMapper.serialize_node().
  """

  def __init__(self):
    self.nodes = []


class FieldName(FieldDecoration):
  """
  A decoration that indicates the name of the field for deserialization and
  serialization. Some serializers may subclass this decoration to allow
  granular overrides and fall back to this class if their subclass is not
  available.
  """

  def __init__(self, name):  # type: (str) -> None
    self.name = name

  def __repr__(self):
    return '{}(name={!r})'.format(type(self).__name__, self.name)


class Format(FieldDecoration):
  """
  A decoration that can be used to annotate the datatype of a field with
  formatting options that should be respected by the (de-) serializer if
  applicable.

  This is used commonly with the #DateType and #DatetimeType, in which case
  the format must be a string or an object with `parse()` and `format()`
  methods.
  """

  def __init__(self, *values):  # type: (Any) -> None
    self.values = values


class InheritKey(FieldDecoration):
  """
  This decoration is used for fields to indicate that during deserialization
  the value should be derived from the key that the #Struct was placed in. The
  datatype for fields with this decoration should be #IntegerType or #StringType.
  """

  def attached(self, field):
    if not isinstance(field.datatype, (IntegerType, StringType)):
      raise TypeError('Field({!r}).datatype should be IntegerType or StringType when decorated '
                      'with Key(), got {} instead.'.format(field.name,
                        field.datatype.to_human_readable()))


class Raw(FieldDecoration):
  """
  This that the field is filled with the raw value of the struct that it is
  deserialized from.
  """


class Remainder(FieldDecoration):
  """
  A decoration for fields to indicate that any unhandled fields during
  deserialization should be stored in that field. The datatype for fields with
  this decoration should be #ObjectType.
  """

  def attached(self, field):
    if not isinstance(field.datatype, ObjectType):
      raise TypeError('Field({!r}).datatype should be ObjectType when decorated with '
                      'Remainder(), got {} instead.'.format(field.name,
                        field.datatype.to_human_readable()))


class StoreNode(GlobalDecoration, ClassDecoration, FieldDecoration):
  """
  This decoration is used to have a deserializer store the #Node instance
  of a #Struct or #Collection in the resulting object's `__databind__`
  metadata under the "node" key.

  Note that this will create a cyclic reference as the #Node references
  the result and the result references the #Node.
  """


class SerializeAs(FieldDecoration, ClassDecoration):
  """
  A class and field decoration to indicate as what type it should be deserialized
  as and serialized to. This works well with interfaces to deserialize them as
  union types where every member of the union type implements said interface.
  """

  def __init__(self, datatype):  # type: (IDataType) -> None
    self.datatype = translate_type_def(datatype)


class Strict(GlobalDecoration, ClassDecoration):
  """
  This class decoration indicates that it should be deserialized in a strict
  fashion, resulting in a deserialization error if any unknown keys are
  encountered.

  The #Strict decoration is inheritable by default.
  """

  def __init__(self, inheritable=True):
    super(Strict, self).__init__(inheritable)


class SkipDefaults(GlobalDecoration, FieldDecoration, ClassDecoration):
  """
  A decoration that indicates values in a #Struct that match the default values
  should be skipped during deserialization.

  The #SkipDefaults decoration is inheritable by default.
  """

  def __init__(self, inheritable=True):
    super(SkipDefaults, self).__init__(inheritable)


class StaticFields(ClassDecoration):
  """
  A decoration for #Struct subclasses that overrides the way struct fields are
  exposed on the calss. When this decoration is present, the fields are exposed
  as their default value rather than the actual #Field object.
  """

  pass


class Validator(FieldDecoration, ClassDecoration):
  """
  A class and field decoration that specifies a function object or method name
  used to validate the deserialized result.
  """

  def __init__(self, function=None, method=None):
    # type: (Optional[Callable[[Any], None]], Optional[str]) -> None
    if (not function and not method) or (function and method):
      raise ValueError('need exactly one of the "function" or "method" arguments')
    self.function = function
    self.method = method

  def test(self, struct_cls, field_value):  # type: (Type[Struct], Any) -> None
    if self.function:
      return self.function(field_value)
    else:
      return getattr(struct_cls, self.method)(field_value)

  @classmethod
  def choices(cls, choices):  # type: (Sequence[Any]) -> Validator
    def _func(value):
      if value not in choices:
        s_choices = '|'.join(map(repr, choices))
        raise ValueError('expected one of {}, got {!r}'.format(s_choices, value))
      return value
    return cls(_func)


class KeywordsOnlyConstructor(ClassDecoration):
  """
  If added to a #Struct subclass, it's constructor will not accept positional arguments.
  This decoration is inheritable by default.
  """

  def __init__(self, inheritable=True):
    super(KeywordsOnlyConstructor, self).__init__(inheritable)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" This module provides a configurable #UnionType for struct fields. """

import importlib
import six
import typing

from nr.collections import abc
from nr.pylang.utils import classdef
from nr.pylang.utils.typing import is_generic, get_generic_args
from nr.interface import attr, default, implements, Interface

from nr.databind.core import IDataType, InvalidTypeDefinitionError
from nr.databind.core.datatypes import PythonClassType, translate_type_def
from nr.databind.core.decorations import SerializeAs
from nr.databind.core.struct import Struct, StructType

__all__ = [
  'IUnionTypeMember',
  'IUnionTypeResolver',
  'StandardTypeResolver',
  'EntrypointTypeResolver',
  'UnionType'
]


class UnknownUnionTypeError(Exception):
  pass


class IUnionTypeMember(Interface):
  """ Represents a member of a #UnionType as returned by a
  #IUnionTypeResolver when given a type name. This interface provides all
  the information about this union member. """

  classdef.comparable(['name', 'type_name', 'datatype'], decorate=default)

  #: The name of the type in the union.
  name = attr()  # type: str

  #: A more easily identifiable name of the type (eg. the full import name
  #: of te type, or an otherwise descriptive name, can be the same as #name).
  type_name = attr()  # type: str

  #: The #IDataType that is used to deserialize this union type member.
  datatype = attr()  # typE: IDataType

  @default
  def isinstance_check(self, value):  # type: (Any) -> bool
    """ Check if *value* is an instance of this union type member. """

    try:
      self.datatype.isinstance_check(value, True, False)
      return True
    except TypeError:
      return False


class IUnionTypeResolver(Interface):
  """ An interface for resolving union types by a name. """

  classdef.comparable([])

  def resolve(self, type_name):  # type: (str) -> IUnionTypeMember
    """ Resolve the *type_name* to a #IUnionTypeMember instance. If the
    *type_name* is unknown, an #UnknownUnionTypeError must be raised. """

  def reverse(self, value):  # type: (Any) -> IUnionTypeMember
    """ Return the #IUnionTypeResolver for the specified *value*, which is
    whatever [[IUnionTypeMember.create_instance()]] returns. Raises
    #UnknownUnionTypeError if the value cannot be reversed. """

  def members(self):  # type: () -> Iterable[IUnionTypeMember]
    """ List up all the members of this resolver. If listing members is not
    supported, a #NotImplementedError must be raised to indicate that. """


@implements(IUnionTypeResolver)
class StandardTypeResolver(object):
  """ This implementation of the #IUnionTypeResolver uses a static mapping
  of type names to #Struct subclasses. It has two forms of initialization:

  1. A (potentially mixed) list of Python type objects, #StructType or
     #PythonClassType objects. In this case the union type name is derived
     from the `__union_type_name__` or `__name__` member (in this order).
  2. A dictionary of union type names mapping to anything that can be
     translated to an #IDataType with #translate_type_def().
  """

  @implements(IUnionTypeMember)
  class _Member(object):
    def __init__(self, name, datatype):
      self.name = name
      self.type_name = datatype.to_human_readable()
      self.datatype = datatype

  classdef.comparable(['types'])

  def __init__(self, types):
    if isinstance(types, abc.Mapping):
      self.types = types
    elif isinstance(types, (list, tuple)):
      self.types = {}
      for item in types:
        if isinstance(item, StructType):
          item = item.struct_cls
        elif isinstance(item, PythonClassType):
          item = item.cls
        elif isinstance(item, type):
          cls = item
        else:
          raise TypeError('expected StructType, PythonClassType or type '
            'object, got {}'.format(type(item).__name__))
        name = getattr(item, '__union_type_name__', item.__class__)
        self.types[name] = item
    else:
      raise TypeError('expected list/tuple/dict, got {}'
                      .format(type(types).__name__))

    for key, value in self.types.items():
      self.register_union_member(key, value, override=True)

  def register_union_member(self, key, value, override=False):
    if not IUnionTypeMember.provided_by(value):
      value = self._Member(key, translate_type_def(value))
    self.types[key] = value

  def resolve(self, type_name):
    try:
      return self.types[type_name]
    except KeyError:
      raise UnknownUnionTypeError(type_name)

  def reverse(self, value):
    result = None
    for key, member in self.types.items():
      try:
        member.datatype.isinstance_check(value, True, False)
      except TypeError:
        continue
      return member
    raise UnknownUnionTypeError(value)

  def members(self):
    try:
      items = self.types.items()
    except NotImplementedError:
      raise NotImplementedError('wrapped "types" mapping does not support iteration')
    return (x[1] for x in items)


@implements(IUnionTypeResolver)
class EntrypointTypeResolver(StandardTypeResolver):
  """ Collects all entries from an entrypoints group. Checks if the class
  loaded via an entrypoint is either a subclass of the specified *base_type*
  or implements it's interface (if *base_type* is a subclass of #Interface).
  """

  @implements(IUnionTypeMember)
  class _Member(object):
    def __init__(self, name, ep, base_type):
      self.name = name
      self._type_name = None
      self._base_type = base_type
      self._ep = ep
      self._datatype = None
      self._cls_cache = None

    @property
    def _cls(self):
      if self._cls_cache is None:
        cls = self._ep.load()
        if not isinstance(cls, type):
          raise TypeError('expected type object for entrypoint {}, got {}'
            .format(self._ep, type(cls).__name__))
        if self._base_type and not issubclass(cls, self._base_type):
          raise TypeError('expected subclasss of {} for entrypoint {}, got {}'
            .format(self._base_type.__name__, self._ep, cls.__name__))
        self._cls_cache = cls
      return self._cls_cache

    @property
    def type_name(self):
      return self._cls.__name__

    @property
    def datatype(self):
      if self._datatype is None:
        self._datatype = translate_type_def(self._cls)
      return self._datatype

  classdef.comparable(['types', 'base_type'])

  def __init__(self, entrypoint_group, base_type=None):
    import pkg_resources
    types = {}
    for ep in pkg_resources.iter_entry_points(entrypoint_group):
      types[ep.name] = self._Member(ep.name, ep, base_type)
    super(EntrypointTypeResolver, self).__init__(types)
    self.base_type = base_type


@implements(IUnionTypeResolver)
class ImportTypeResolver(object):
  """ This type resolver identifies a union type by their fully qualified
  Python import name, constructed from the `__module__` and `__name__`
  attributes of a type. """

  classdef.comparable([])

  def resolve(self, type_name):  # type: (str) -> IUnionTypeMember
    module_name, member = type_name.rpartition('.')[::2]
    if six.PY2 and module_name == 'builtins':
      module_name = '__builtin__'
    try:
      module = importlib.import_module(module_name)
    except ImportError:
      raise UnknownUnionTypeError(type_name)
    try:
      cls = getattr(module, member)
    except AttributeError:
      raise UnknownUnionTypeError(type_name)
    if not isinstance(cls, type):
      raise UnknownUnionTypeError(type_name)
    try:
      datatype = translate_type_def(cls)
    except InvalidTypeDefinitionError:
      raise UnknownUnionTypeError(type_name)
    return IUnionTypeMember(name=type_name, type_name=type_name,
      datatype=datatype, isinstance_check=lambda x: isinstance(x, cls))

  def reverse(self, value):
    module_name = type(value).__module__
    if six.PY2 and module_name == '__builtin__':
      module_name = 'builtins'
    member = type(value).__name__
    type_name = module_name + '.' + member
    return IUnionTypeMember(
      name=type_name, type_name=type_name,
      datatype=translate_type_def(type(value)),
      isinstance_check=lambda x: isinstance(x, type(value)))

  def members(self):
    raise NotImplementedError


@implements(IDataType)
class UnionType(object):
  """ The UnionType represents multiple types. A value represented by this
  datatype can be of any of the types that are encapsulated by the union
  type. UnionType only supports the encapsulation of #StructTypes.

  The UnionType can operate in two modes for the serialization and
  deserialization. In either mode, the object from which the UnionType is
  deserialized must contain a "type" key (configurable with the `type_key`
  parameter).

  In the default mode, the fields for the type are read from the same level.

  ```yaml
  type: onePossibleUnionType
  someField: value
  ```

  With the `nested` option enabled, the values are instead read from an object
  nested with the same name as the type.

  ```yaml
  type: onePossibleUnionType
  onePossibleUnionType:
    someField: value
  ```

  The #StandardTypeResolver is used in the usual case.

  Union type can be conveniently defined using lists with more than one item,
  the [[typing.Union]] type or dictionaries. In case of a dictionary, the
  union type name is defined in the dictionary key. Otherwise, it is read
  from the `__union_type_name__` or classname.

  # Example: Subclasses as Union type members

  A common use case for union types is the registration on a common base type, where
  the base type only acts as a proxy for all the possible subtypes.

  ```yaml
  from nr.databind.core import Field, SerializeAs, Struct, UnionType

  @SerializeAs(UnionType())
  class Pet(Struct):
    ...

  @UnionType.extend(Pet, 'cat')
  class Cat(Pet):
    fur_color = Field(str)

  @UnionType.extend(Pet, 'dog')
  class Dog(Pet):
    bark_volume = Field(int)
  ```

  # Example: Union types over entrypoints

  ```yaml
  from nr.databind.core import UnionType, SerializeAs
  from nr.interface import Interface

  @SerializeAs(UnionType.entrypoint('my_package.plugins.IMyPlugin')
  class IMyPlugin(Interface):
    def foo(self):
      ...
  ```

  This allows plugins to be defined in Python package entrypoints, pointing to the class
  that implements the #Interface. Note that this subclass must inherit from the #Struct
  class in order to be deserializable.

  ```yaml
  from nr.databind.core import Field, Struct
  from nr.interface import implements, override
  from my_package.plugins import IMyPlugin

  @implements(IMyPlugin)
  class ThePlugin(Struct):
    option_1 = Field(str)
    option_2 = Field(int)

    @override
    def foo(self):
      # ...
  ```
  """

  #: Import these members on the UnionType to reduce the number of
  #: imports that need to be made when implementing a custom type resolver
  #: or creating one of the standard implementations.
  UnknownUnionTypeError = UnknownUnionTypeError
  ITypeMember = IUnionTypeMember
  ITypeResolver = IUnionTypeResolver
  StandardTypeResolver = StandardTypeResolver
  EntrypointTypeResolver = EntrypointTypeResolver
  ImportTypeResolver = ImportTypeResolver

  classdef.comparable(['type_resolver', 'type_key', 'nested'])

  def __init__(self, type_resolver=None, type_key='type', nested=False):
    """
    # Arguments
    type_resolver (IUnionTypeResolver, dict): The type resolver for this union type,
      or a dictionary that will be passed to the #StandardTypeResolver. If no argument
      is provided, an empty #StandardTypeResolver will be created that can be extended
      with the #UnionType.extend() method.
    type_key (str): The key in the payload that identifies the union type member.
    nested (bool): If `True`, the payload to deserialize the union type member from are
      nested in a key with the name of the union member. Otherwise, the remaining keys in
      the union payload are forwarded (note that in this case the *type_key* can not overlap
      with a field in the union member).
    """

    if type_resolver is None:
      type_resolver = StandardTypeResolver({})
    if isinstance(type_resolver, (dict, list, tuple)):
      type_resolver = StandardTypeResolver(type_resolver)

    self.type_resolver = type_resolver
    self.type_key = type_key
    self.nested = nested

  @classmethod
  def with_entrypoint_resolver(cls, *args, **kwargs):
    """ Deprecated. Use #entrypoint() instead. """

    return cls.entrypoint(*args, **kwargs)

  @classmethod
  def with_import_resolver(cls, *args, **kwargs):
    """ Deprecated. Use #importing() instead. """

    return cls.importing(*args, **kwargs)

  @classmethod
  def entrypoint(cls, *args, **kwargs):
    """
    Returns a #UnionType instance initialized with an #EntrypointTypeResolver from
    the specified *args* and *kwargs*.

    # Arguments
    args (Any): Positional arguments forwarded to #EntrypointTypeResolver.
    type_key (str): The *type_key* argument forwarded to the #UnionType constructor.
    nested (bool): The *nested* argument forwarded to the #UnionType constructor.
    kwargs (Any) Keyword arguments forwarded to #EntrypointTypeResolver.
    """

    union_kwargs = {
      k: kwargs.pop(k) for k in ('type_key', 'nested') if k in kwargs}
    return cls(EntrypointTypeResolver(*args, **kwargs), **union_kwargs)

  @classmethod
  def importing(cls, *args, **kwargs):
    """
    Returns a #UnionType instance initialized with an #ImportTypeResolver from the
    specified *args* and *kwargs*.

    # Arguments
    args (Any): Positional arguments forwarded to #ImportTypeResolver.
    type_key (str): The *type_key* argument forwarded to the #UnionType constructor.
    nested (bool): The *nested* argument forwarded to the #UnionType constructor.
    kwargs (Any) Keyword arguments forwarded to #ImportTypeResolver.
    """

    union_kwargs = {
      k: kwargs.pop(k) for k in ('type_key', 'nested') if k in kwargs}
    return cls(ImportTypeResolver(*args, **kwargs), **union_kwargs)

  @classmethod
  def extend(cls, union, name, member=None):
    # type: (Union[UnionType, Type], str, Optional[Type[Struct]]) -> Any
    """
    This is a utility function to add a member to a #UnionType that harbours a
    #StandardTypeResolver. This is useful particularly when first declaring a class
    that serializes as a #UnionType and only later declares it's union members (see the
    #UnionType Pet example).

    # Arguments
    union (UnionType, type): Either a #UnionType object or a type object that is decorated
      with #SerializeAs, which in turn contains the #UnionType. The union must harbour a
      #StandardTypeResolver.
    name (str): The name of the new union member.
    member (Type[Struct]): The member to place into the #StandardTypeResolver. If this is
      not set, a decorator will be returned that can be used to decorate a class definition.
    """

    if not IDataType.provided_by(union):
      dec = SerializeAs.get(union)
      if not dec:
        raise RuntimeError('no SerailizeAs decoration found on {!r}'.format(union))
      union = dec.datatype
    if not isinstance(union, UnionType):
      raise RuntimeError('expected UnionType, found {}'.format(type(union).__name__))
    if not isinstance(union.type_resolver, StandardTypeResolver):
      raise RuntimeError('expected StandardTypeResolver in union, found {}'.format(
        type(union.type_resolver).__name__))

    if member is None:
      def decorator(cls):
        union.type_resolver.register_union_member(name, cls)
        return cls
      return decorator
    else:
      union.type_resolver.register_union_member(name, cls)
      return None

  # IDataType

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    if is_generic(py_type_def, typing.Union):
      union_types = get_generic_args(py_type_def)
      if all(issubclass(x, Struct) for x in union_types):
        return UnionType(union_types)
    elif isinstance(py_type_def, list) and len(py_type_def) > 1:
      return UnionType([recursive(x) for x in py_type_def])
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):
    try:
      members = list(self.type_resolver.members())
    except NotImplementedError:
      # TODO (@NiklasRosenstein): We use check_value() to type check when
      #   initializing a Struct instance's field, and we don't want that to
      #   fail just because the type resolver doesn't support member listing.
      #   Maybe make this behavior configurable?
      return py_value
    if not any(x.isinstance_check(py_value) for x in members):
      raise TypeError('expected {{{}}}, got {}'.format(
        '|'.join(sorted(x.name for x in members)),
        type(py_value).__name__))
    return py_value



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import copy
import pkg_resources
import six
import typing

from .decorations import KeywordsOnlyConstructor
from nr.collections import abc, OrderedDict
from nr.databind.core import IDataType, InvalidTypeDefinitionError
from nr.databind.core.datatypes import CollectionType, translate_type_def
from nr.databind.core.decorations import FieldDecoration, StaticFields
from nr.pylang.utils import NotSet
from nr.pylang.utils import classdef, funcdef
from nr.pylang.utils.typing import is_generic, get_generic_args, extract_optional
from nr.interface import implements
from nr.stream import Stream

__all__ = [
  'Field',
  'FieldSpec',
  'Struct',
  'StructType',
  'make_struct'
]


@implements(IDataType)
class StructType(object):
  """ Represents the datatype for a [[Struct]] subclass. """

  classdef.comparable(['struct_cls', 'ignore_keys'])
  _INLINE_GENERATED_TYPENAME = '_InlineStructAdapter__generated'

  def __init__(self, struct_cls, ignore_keys=None):
    assert isinstance(struct_cls, type), struct_cls
    assert issubclass(struct_cls, Struct), struct_cls
    self.struct_cls = struct_cls
    self.ignore_keys = ignore_keys or []

  def to_human_readable(self):
    return self.struct_cls.__name__

  def propagate_field_name(self, name):
    if self.struct_cls.__name__ == self._INLINE_GENERATED_TYPENAME:
      self.struct_cls.__name__ = name.rpartition('.')[-1]
      self.struct_cls.__qualname__ = name

  @classmethod
  def from_typedef(cls, recursive, py_type_def):
    # Struct subclass
    if isinstance(py_type_def, type) and issubclass(py_type_def, Struct):
      return cls(py_type_def)
    # Inline definition
    if isinstance(py_type_def, dict):
      return cls(type(cls._INLINE_GENERATED_TYPENAME, (Struct,), py_type_def))
    raise InvalidTypeDefinitionError(py_type_def)

  def isinstance_check(self, py_value, strict, coerce):  # type: (Any, bool, bool) -> Any
    if (strict and type(py_value) != self.struct_cls) or \
       (not strict and not isinstance(py_value, self.struct_cls)):
      raise TypeError('expected {} instance, got {}'.format(
        self.struct_cls.__name__, type(py_value).__name__))
    return py_value


class Field(object):
  """ Base class for struct fields. The *datatype* field will be translated
  to an #IDatatype instance when it is processed by the #_StructMeta
  class.

  Any positional arguments beyond the field's data-type must be #Decoration
  objects. Other arguments must be supplied as keyword-arguments. """

  classdef.repr('name datatype nullable default')
  classdef.comparable('__class__ name datatype nullable default')
  _INSTANCE_INDEX_COUNTER = 0

  DEFAULT_CONSTRUCT = classdef.make_singleton('Field.DEFAULT_CONSTRUCT')

  def __init__(self, datatype, *decorations, **kwargs):
    """
    :param datatype: The datatype for the field. The value of this argument is
      automatically converted to an #IDataType object with #translate_type_def().
    :param decorations: Zero or more #FieldDecoration objects.
    :param name: The name of the field. When a field is inserted into a #FieldSpec,
      the name will be set automatically. If the name is already set at that point
      in time, the name of the field must match the name that the field would be
      assigned to in the #FieldSpec.
    :param default: The default value of the field. If the passed value is callable,
      it will be called to retrieve the actual default value on demand. It can also
      be #Field.CONSTRUCT in which case the *datatype* must be a #StructType or
      #CollectionType.
    :param nullable: Whether the field can have #None as a value. If the *default*
      value of the field is set to #None, this will be #True automatically. Otherwise
      it will default to #False.
    :param required: Whether the field is required to be present during (de-)
      serialization. Defaults to #False.
    :param hidden: If this is set to True, the field is disregarded during (de-)
      serialization and serves as a slot that is available at runtime only.
    :param prominent: Set to #True if the field is a "prominent" field, eg. it should
      be included in #Struct.__repr__().
    """

    name = kwargs.pop('name', None)
    assert name is None or isinstance(name, str), repr(name)
    nullable = kwargs.pop('nullable', None)
    default = kwargs.pop('default', NotSet)
    required = kwargs.pop('required', False)
    hidden = kwargs.pop('hidden', False)
    prominent = kwargs.pop('prominent', False)
    funcdef.raise_kwargs(kwargs)

    if is_generic(datatype, typing.Optional):
      datatype = get_generic_args(datatype)[0]
      if nullable is not None:
        # For consistency reasons we don't allow "Field(Optional[...], nullable=True)"
        # as it looks confusing.
        raise ValueError('"nullable" argument must be None when an Optional is provided '
                         'to the Field() constructor')
      nullable = True

    if default is None:
      if nullable is None:
        nullable = True
      elif not nullable:
        raise ValueError('default=None but nullable=False, this does not add up')

    if not IDataType.provided_by(datatype):
      datatype = translate_type_def(datatype)

    for item in decorations:
      if not isinstance(item, FieldDecoration):
        raise TypeError('expected FieldDecoration object, got {!r}'.format(type(item).__name__))

    if default is Field.DEFAULT_CONSTRUCT and type(datatype) != StructType:
      raise ValueError('default cannot be Field.DEFAULT_CONSTRUCT if the '
        'field datatype is not a StructType (got {})'.format(
          type(datatype).__name__))

    if default is NotSet and hidden:
      raise ValueError('a hidden field requires a default value')

    self.datatype = datatype
    self.decorations = list(decorations)
    self.name = name
    self.nullable = False if nullable is None else bool(nullable)
    self.default = default
    self.required = required
    self.hidden = hidden
    self.prominent = prominent
    self.parent = None

    self.instance_index = Field._INSTANCE_INDEX_COUNTER
    Field._INSTANCE_INDEX_COUNTER += 1

    for item in decorations:
      item.attached(self)

  def __call__(self, *args, **kwargs):
    """ Syntactic sugar for fields that wrap a #StructType. Creates an
    instance of the wrapped type. """

    if isinstance(self.datatype, StructType):
      return self.datatype.struct_cls(*args, **kwargs)
    raise RuntimeError('{} cannot be called as it does not wrap a StructType'
      .format(self))

  def get_priority(self):  # type: () -> int
    """ The priority determines when the field will have its chance to
    extract values from the source dictionary. The default priority
    is zero (0). """

    return 0

  def set_name(self, name):  # type: str -> None
    """ Sets the name of the field. if the name is already set, a
    #RuntimeError will be raised. """

    if self.name is not None:
      raise RuntimeError('cannot set field name to {!r}, name is already '
                         'set to {!r}'.format(name, self.name))
    if not isinstance(name, str):
      raise TypeError('Field.name must be a string, got {}'
                      .format(type(name).__name__))
    self.name = name

  def get_struct_class_member(self, struct_cls):  # type: (Type[Struct]) -> Any
    """ This method is called when the field is accessed via
    [[StructMeta.__getattr__()]] and can be used to expose a class-level
    property on the #Struct class.

    Return #NotSet if no property is to be exposed.

    The default implementation checks if the [[.datatype]] is an instance of
    #StructType and returns the wrapped #Struct subclass in that case.
    """

    if isinstance(self.datatype, StructType):
      return self.datatype.struct_cls
    return NotSet

  def get_default_value(self):
    if self.default is NotSet:
      raise RuntimeError('Field({!r}).default is NotSet'.format(self.name))
    if callable(self.default):
      return self.default()
    if self.default is Field.DEFAULT_CONSTRUCT:
      return self.datatype.struct_cls()
    return self.default

  def check_value(self, struct_name, py_value, strict=False, coerce=True):
    try:
      return self.datatype.isinstance_check(py_value, strict, coerce)
    except (TypeError, ValueError) as exc:
      raise type(exc)('{}.{}: {}'.format(struct_name, self.name, exc))


class FieldSpec(object):
  """
  A container for #Fields which is used to collect all fields of a #Struct in
  a single place.
  """

  classdef.comparable('_fields')

  def __init__(self, fields=None):
    """ Creates a new #FieldSpec object from a list of #Field
    objects. Note that all fields must have a name, otherwise a #ValueError
    is raised. """

    fields = list(fields or [])
    for field in fields:
      if not isinstance(field, Field):
        raise TypeError('expected Field, got {!r}'
                        .format(type(field).__name__))
      if not field.name:
        raise ValueError('found unnamed field: {!r}'.format(field))
      assert isinstance(field.name, str), field

    fields.sort(key=lambda x: x.instance_index)

    self._fields = OrderedDict((x.name, x) for x in fields)
    self._update_cache()

  def _update_cache(self):
    self._fields_indexable = list(self._fields.values())
    self._fields_prominent = [f for f in self._fields_indexable if f.prominent]

  @classmethod
  def from_dict(cls, fields_dict):
    """ Compiles a #FieldSpec from a dictionary that is expected to consist
    of only field definitions. Adapts Python type definitions to standard
    #Field objects. #Fields in this dictionary that don't have a
    name will be assigned the name of their associated key. """

    fields = []
    for key, field in six.iteritems(fields_dict):
      if not isinstance(field, Field):
        field = Field(field)
      if not field.name:
        field.name = key
      fields.append(field)
    return cls(fields)

  @classmethod
  def from_annotations(cls, obj_class):
    """ Compiles a #FieldSpec object from the class member annotations in
    the class *obj_class*. The annotation value is the field's datatype.
    If a value is assigned to the class member, it acts as the default value
    for that field.

    Type annotations can be wrapped in the #Optional generic to indicate
    that the field is nullable. Alternatively, the default value of the field
    can be set to `None`. """

    fields = []
    for name, datatype in six.iteritems(obj_class.__annotations__):
      default = getattr(obj_class, name, NotSet)
      field = Field(
        datatype=datatype,
        default=default,
        name=name)
      fields.append(field)
    return cls(fields)

  @classmethod
  def from_class_members(cls, obj_class):
    """ Compiles a #FieldSpec object from the #Field members on a class. """

    fields = []
    for name, value in six.iteritems(vars(obj_class)):
      if not isinstance(value, Field):
        continue
      if not value.name:
        value.name = name
      elif value.name != name:
        raise RuntimeError('mismatched field name {!r} != {!r}'
                           .format(value.name, name))
      fields.append(value)
    return cls(fields)

  @classmethod
  def from_list_def(cls, list_def):
    """ Compiles a FieldSpec from a list of tuples. Every tuple must have at
    least two elements, the first defining the name of the field, the second
    the type. An optional third field in the tuple may be used to specify
    the field default value. """

    assert not isinstance(list_def, abc.Mapping), "did not expect a mapping"

    if isinstance(list_def, str):
      if ',' in list_def:
        list_def = [x.strip() for x in list_def.split(',')]
      else:
        list_def = list_def.split()

    fields = []
    for item in list_def:
      if isinstance(item, str):
        field = Field(object, name=item)
      elif isinstance(item, tuple):
        name, datatype = item[:2]
        default = item[2] if len(item) > 2 else NotSet
        field = Field(datatype, default=default, name=name)
        fields.append(field)
      elif isinstance(item, Field):
        if not item.name:
          raise ValueError('unbound field in __fields__ list')
        field = item
      else:
        raise TypeError('expected {str, tuple, Field}, got {!r}'
                        .format(type(item).__name__))
      fields.append(field)
    return cls(fields)

  def __getitem__(self, name):
    return self._fields[name]

  def __setitem__(self, name, field):  # type: (str, Field)
    if not field.name:
      field.name = name
    self.update([field])

  def __delitem__(self, name):
    del self._fields[name]
    self._update_cache()

  def __contains__(self, name):
    return name in self._fields

  def __iter__(self):
    return six.iterkeys(self._fields)

  def __len__(self):
    return len(self._fields)

  def __repr__(self):
    return 'FieldSpec({!r})'.format(list(self._fields.values()))

  def __copy__(self):
    return type(self)(self._fields.values())

  def keys(self):  # type: () - >Stream[str]
    return Stream(six.iterkeys(self._fields))

  def values(self):  # type: () -> Stream[Field]
    return Stream(six.itervalues(self._fields))

  def items(self):  # type: () -> Stream[Tuple[str, Field]]
    return Stream(six.iteritems(self._fields))

  def update(self, fields):
    # type: (FieldSpec) -> FieldSpec
    """ Updates this #FieldSpec with the files from another spec and returns
    *self*. This operation maintains the order of existing fields in the spec.
    """

    if not isinstance(fields, FieldSpec):
      fields = FieldSpec(fields)

    for key, value in fields._fields.items():
      self._fields[key] = value
    self._update_cache()
    return self

  def get(self, key, default=None):
    return self._fields.get(key, default)

  def get_index(self, index):
    # type: (int) -> Field
    return self._fields_indexable[index]

  def has_prominent_fields(self):  # type: () -> bool
    return bool(self._fields_prominent)

  def prominent(self):  # type: () -> Iterable[Field]
    return iter(self._fields_prominent)

  def copy(self):
    return self.__copy__()


class _StructMeta(type):
  """ Private. Meta class for #Struct. """

  def __init__(self, name, bases, attrs):
    # Collect inherited fields.
    parent_fields = FieldSpec()
    for base in bases:
      if hasattr(base, '__fields__') and isinstance(base.__fields__, FieldSpec):
        parent_fields.update(base.__fields__)

    # If there are any class member annotations, we derive the object fields
    # from these rather than from class level #Field objects.
    if '__fields__' in attrs and not isinstance(attrs['__fields__'], FieldSpec):
      fields = FieldSpec.from_list_def(self.__fields__)
    elif '__annotations__' in attrs:
      if isinstance(attrs['__annotations__'], dict):
        assert attrs['__annotations__'] is self.__annotations__
        fields = FieldSpec.from_annotations(self)
      else:
        fields = FieldSpec.from_list_def(attrs['__annotations__'])
    else:
      fields = FieldSpec.from_class_members(self)
      if not fields and '__fields__' in attrs:
        fields = attrs['__fields__']

    # If at the class-level a field has been overwritten with a plain value,
    # it will cause a copy of the field to be created and the default value
    # to be overwritten.
    for key, field in parent_fields.items():
      assert field.name == key
      value = attrs.get(field.name, field)
      if not isinstance(value, Field) and value != field.default:
        new_field = copy.copy(field)
        new_field.default = value
        new_field.parent = field
        parent_fields.update([new_field])

    # If a struct is declared with __fields__ or by directly setting
    # __annotations__, the default value may still be defined on the class
    # level.
    for key, field in fields.items():
      assert field.name == key
      value = attrs.get(field.name, field)
      if value is not field and value != field.default:
        if field.default is not NotSet:
          raise RuntimeError('cannot override default value for field {!r} '
            'when the field was declared with a default value ({!r}) on the '
            'same class'.format(key, field.default))
        field.default = value

    # Give new fields (non-inherited ones) a chance to propagate their
    # name (eg. to datatypes, this is mainly used to automatically generate
    # a proper class name for inline-declared objects).
    for field in fields.values():
      field.datatype.propagate_field_name(self.__name__ + '.' + field.name)

    fields = parent_fields.update(fields)
    for key in fields:
      if key in vars(self):
        delattr(self, key)

    self.__has_static_fields = None
    self.__fields__ = fields

  def __getattr__(self, name):
    field = self.__fields__.get(name)

    if self.__has_static_fields is None:
      self.__has_static_fields = bool(StaticFields.get(self))

    if field and self.__has_static_fields and field.default is not NotSet:
      return field.default

    if field and not self.__has_static_fields:
      if isinstance(field.datatype, StructType):
        return field.datatype.struct_cls
      else:
        return field

    raise AttributeError('{} has no attribute {}'.format(self.__name__, name))


class Struct(six.with_metaclass(_StructMeta)):
  """
  An object is comprised of field descriptors and metadata which are used to
  build the object from a nested structure. Objects can be defined in two
  major ways: With the #Field class, or with the class member annotation
  syntax that is available since Python 3.6.

  With annotations:

  ```py
  from typing import Optional
  class Person(Struct):
    name: str
    age: Optional[int]
    telephone_numbers: [str] = lambda: []
  ```

  With the #Field class:

  ```py
  class Person(Struct):
    name = Field(str)
    age = Field(str, optional=True)
    telephone_numbers = Field([str], default=lambda: [])
  ```

  Both objects show the same semantics and can be deserialized from a
  this example YAML data:

  ```yaml
  people:
    - name: Barbara
    - name: John
      telephone_numbers:
        - "1 432 9876543"
    - name: Will
      age: 52
      telephone_numbers:
        - "1 234 5678912"
  ```
  """

  __fields__ = FieldSpec()

  def __init__(self, *args, **kwargs):
    struct_name = type(self).__name__
    argcount = len(args) + len(kwargs)
    if argcount > len(self.__fields__):
      # TODO(nrosenstein): Include min number of args.
      raise TypeError('{}() constructor expected at max {} arguments, got {}'
                      .format(type(self).__name__, len(self.__fields__), argcount))

    if KeywordsOnlyConstructor.get(type(self)) and args:  # TODO: Caching?
      raise TypeError('{}() does not accept positional arguments'.format(type(self).__name__))

    # Check the type of all keyword arguments.
    for key, value in kwargs.items():
      try:
        field = self.__fields__[key]
      except KeyError:
        raise TypeError('{}() constructor received unexpected keyword argument {!r}'
                        .format(type(self).__name__, key))
      if value is None and field.nullable:
        continue
      kwargs[key] = field.check_value(struct_name, value)

    # Add all arguments to the kwargs for extraction.
    for field, arg in zip(self.__fields__.values(), args):
      if field.name in kwargs:
        raise TypeError('duplicate arguments for "{}"'.format(field.name))
      if arg is None and field.nullable:
        kwargs[field.name] = None
        continue
      try:
        kwargs[field.name] = field.check_value(struct_name, arg)
      except TypeError as exc:
        raise TypeError('{}.{}: {}'.format(type(self).__name__, field.name, exc))

    # Extract all fields.
    handled_keys = set()
    missing_keys = set()
    for field in self.__fields__.values().sortby(lambda x: x.get_priority()):
      if field.name not in kwargs and field.name not in vars(self):
        if field.default is NotSet:
          missing_keys.add(field.name)
          continue
        kwargs[field.name] = field.get_default_value()
      handled_keys.add(field.name)

    # Give the class a chance to return the defaults for missing keys.
    if missing_keys:
      defaults = self._get_field_defaults(missing_keys)
      defaults.update(kwargs)
      kwargs = defaults

    for key in missing_keys:
      if key not in kwargs:
        raise TypeError('{}: missing required argument "{}"'.format(
          type(self).__name__, field.name))

    unhandled_keys = set(kwargs.keys()) - handled_keys
    if unhandled_keys:
      raise TypeError('unexpected keyword arguments: {!r}'.format(unhandled_keys))

    vars(self).update(kwargs)
    self.__databind__ = {}

  def __eq__(self, other):
    if type(other) != type(self):
      return False
    for key in self.__fields__:
      if getattr(self, key) != getattr(other, key):
        return False
    return True

  def __ne__(self, other):
    if type(other) != type(self):
      return True
    for key in self.__fields__:
      if getattr(self, key) != getattr(other, key):
        return True
    return False

  def __repr__(self):
    if self.__fields__.has_prominent_fields():
      fields = self.__fields__.prominent()
    else:
      fields = self.__fields__.values()
    attrs = ['{}={!r}'.format(f.name, getattr(self, f.name)) for f in fields]
    return '{}({})'.format(type(self).__name__, ', '.join(attrs))

  @classmethod
  def _get_field_defaults(cls, field_subset):  # type; (Set[str]) -> Dict[str, Any]
    """
    Class method that may be overwritten to provide default values when the Struct instance
    is created that is not described in the #Field.default of the struct fields.
    """

    return {}


def make_struct(name, fields, base=None, mixins=()):
  """
  Creates a new #Struct subclass with the specified fields. The fields must
  be a dictionary of bound #Field objects or a dictionary of unbound ones.

  >>> from nr.databind.core.struct import struct
  >>> Person = struct('name,age,address')
  >>> Person('Alexa Smith', 42, None)
  """

  if isinstance(fields, str):
    if ',' in fields:
      fields = [x.strip() for x in fields.split(',')]
    else:
      fields = fields.split()

  if isinstance(fields, abc.Mapping):
    fields = FieldSpec.from_dict(fields)
  else:
    fields = FieldSpec.from_list_def(fields)

  if base is None:
    base = Struct

  for key, value in fields.items():
    if not isinstance(key, str):
      raise TypeError('class member name must be str, got {}'
                      .format(type(key).__name__))

  return type(name, (base,) + mixins, {'__fields__': fields})


def cast_struct(src, target_cls, **update):  # type: (Struct, Type[Struct], Any) -> Struct
  """
  Cast a struct instance from one to another class by taking all overlapping properties
  from *src* and passing it to the *target_cls* constructor.
  """

  kwargs = {k: getattr(src, k) for k in src.__fields__ if k in target_cls.__fields__}
  kwargs.update(update)
  return target_cls(**kwargs)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core.datatypes import CollectionType, translate_type_def
import six


class _CollectionMeta(type):

  def __new__(cls, name, bases, attrs):
    item_type = attrs.get('item_type', object)
    attrs['item_type'] = translate_type_def(item_type)
    return super(_CollectionMeta, cls).__new__(cls, name, bases, attrs)

  @property
  def datatype(cls):  # type: () -> IDataType
    return CollectionType(cls.item_type, cls)


class Collection(six.with_metaclass(_CollectionMeta)):
  """
  Base class for defining a custom collection of items. Using this class
  allows you to define an array/list datatype while adding any attributes or
  methods to it.

  Subclassing [[Collection]] must always be combined with an actual collection
  implementation ([[list]], [[set]], [[collections.deque]], etc.)

  Example:

  ```py
  from nr.databind import Collection, ObjectMapper
  from nr.databind.json import JsonModule

  class Items(list, Collection):
    item_type = str

    def getstring(self):
      return ''.join(self)

  items = ObjectMapper(JsonModue()).deserialize(['a', 'b', 'c'], Items)
  assert items.getstring() == 'abc'
  ```
  """

  def __init__(self, *args, **kwargs):
    super(Collection, self).__init__(*args, **kwargs)
    self.__databind__ = {}
